@@ -1,4 +1,4 @@
-import websocket
+import websocket #upm package(websocket-client)
 import threading
 import time
 import requests
@@ -16,11 +16,12 @@ from hashlib import sha256
 
 '''A BETTER Stack Exchange chat module.'''
 
+
 class Room:
-  def __init__(self, parent, roomID, logRequestErrors = False):
-    '''A room which the bot is in. You should never create this classs by hand!'''
+  '''A room which the bot is in. You should never create this classs by hand!'''
+  def __init__(self, parent, roomID, logRequestErrors = False, autoConnect = True):
     self.session = parent.session
-    self.fkey = parent.fkey
+    self._fkey = parent.fkey
     self.userID = parent.userID
     self.roomID = roomID
     self.logRequestErrors = logRequestErrors
@@ -30,14 +31,15 @@ class Room:
     self.thread = None
     self.socket = None
     self.running = False
-    self.handlers = {}
+    self.handlers = {i.value: set() for i in Events}
     self.internalHandlers = {
       Events.REPLY.value: self._replyHandler,
       Events.MENTION.value: self._replyHandler
     }
     self.lastPing = 0
 
-    self.connect()
+    if autoConnect:
+      self.connect()
 
   def __del__(self):
     self.halt()
@@ -60,12 +62,13 @@ class Room:
       r = self.session.post(
         "https://chat.stackexchange.com/ws-auth",
         data = {
-          "fkey": self.fkey,
+          "fkey": self._fkey,
           "roomid": self.roomID
         }
-      ).json()
+      )
+      r = r.json()
     except Exception as e:
-      raise errors.ConnectionError("Unable to authenticate") from e
+      raise errors.ConnectionError("Unable to authenticate: " + str(r.content)) from e
     target = r["url"]+"?l={}".format(int(time.time()))
     try:
       self.socket = websocket.create_connection(target, origin="http://chat.stackexchange.com", timeout = 2)
@@ -98,7 +101,7 @@ class Room:
           "since": since,
           "mode": mode,
           "count": count,
-          "fkey": self.fkey
+          "fkey": self._fkey
         },
         headers = {
           'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -122,14 +125,27 @@ class Room:
         continue
       except (websocket.WebSocketConnectionClosedException, ConnectionResetError):
         self.logger.warning("Connection closed, attempting to reconnect")
+        self.socket.close()
+        time.sleep(2)
         self.connect()
       except Exception as e:
+        self.logger.info("Shutting down...")
+        self.socket.close()
+        self.session.post(
+          "https://chat.stackexchange.com/chats/leave/"
+          + str(self.roomID),
+          data = {
+            "quiet": True,
+            "fkey": self._fkey
+          }
+        )
+        self.running = False
         self.logger.critical("Unexpected error")
         raise errors.ConnectionError from e
       if data is not None and data != "":
         try:
           data = json.loads(data)
-        except json.JSONDecodeError, TypeError:
+        except (json.JSONDecodeError, TypeError):
           self.logger.warning("Recived corrupted data: " + data)
         else:
           self.lastPing = time.time()
@@ -144,7 +160,7 @@ class Room:
       + str(self.roomID),
       data = {
         "quiet": True,
-        "fkey": self.fkey
+        "fkey": self._fkey
       }
     )
     self.socket.close()
@@ -169,36 +185,38 @@ class Room:
       :param callback: The callback function. Gets a namedtuple with all the event data as its only parameter.
       :type callback: function
 
-      :raises ValueError: If there is an event handler already registered for that event type or the event type is unknown.
+      :raises ValueError: If the event type is unknown.
     '''
     if event in Events:
-      if event.value not in self.handlers:
-        self.handlers[event.value] = callback
-      else:
-        raise ValueError("Handler already registered for event " + Events(event).name)
+        self.handlers[event.value].add(callback)
     else:
       raise ValueError("Unknown event type: " + str(event))
 
-  def off(self, event):
+  def off(self, listener):
     '''Remove an event listener.
 
-      :param event: The event type of the handler to remove.
-      :type event: sechat.events.EventType
+      :param listener: The function to remove.
+      :type listener: function
     '''
-    if event in Events:
-      if event.value in self.handlers:
-        self.handlers.pop(event.value)
-      else:
-        raise ValueError("No handler for event " + Events(event).name)
+    toRemove = None
+    for handlers in self.handlers:
+      for handler in self.handlers[handlers]:
+        if handler == listener:
+          toRemove = handlers
+          break
+    if toRemove is not None:
+      self.handlers[toRemove].remove(listener)
     else:
-      raise ValueError("Unknown event type: " + str(event))
+      raise ValueError("Listener not registered")
+          
 
   def handle(self, event, data, default = None):
     if event in self.internalHandlers:
       self.internalHandlers[event](data)
     if event in self.handlers:
       t = namedtuple("Event", data.keys())
-      self.handlers[event](t(**data))
+      for h in self.handlers[event]:
+        h(t(**data))
     elif default:
       default(data)
 
@@ -210,7 +228,7 @@ class Room:
       "https://chat.stackexchange.com/messages/ack",
       data = {
         "id": event["message_id"],
-        "fkey": self.fkey
+        "fkey": self._fkey
       }
     )
 
@@ -222,7 +240,7 @@ class Room:
       if self.logRequestErrors:
         self.logger.exception("An error occured in function " + repr(func))
       return
-    if r.text.startswith("You can perform this action again"):
+    if r.text.startswith("You can perform"):
       if handleTooFast:
         time.sleep(self.cooldown)
         self.cooldown = self.cooldown ** 2
@@ -230,7 +248,8 @@ class Room:
       else:
         raise errors.TooFastError
     elif r.text.startswith("It is too late"):
-      return False
+      self.cooldown = 2
+      raise errors.TimeoutError
     else:
       self.cooldown = 2
       return r
@@ -256,7 +275,7 @@ class Room:
         "firstMessageId": start,
         "lastMessageId": end,
         "title": title,
-        "fkey": self.fkey
+        "fkey": self._fkey
       },
       headers = {
         'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -278,7 +297,7 @@ class Room:
       "https://chat.stackexchange.com/conversation/delete/{0}/{1}"
       .format(self.roomID, title),
       data = {
-        "fkey": self.fkey
+        "fkey": self._fkey
       }
     )
       
@@ -293,13 +312,13 @@ class Room:
       :return: The ID of the message that was just sent.
       :rtype: int
     '''
-    self.logger.info("Sending message: " + message)
-    return self.processTooFast(
+    self.logger.info("Sending message: " + message) 
+    r = self.processTooFast(
         lambda: self.session.post(
           "https://chat.stackexchange.com/chats/{}/messages/new"
             .format(self.roomID),
           data = {
-            "fkey": self.fkey,
+            "fkey": self._fkey,
             "text": message
           },
           headers = {
@@ -309,7 +328,11 @@ class Room:
           }  
       ),
       handleTooFast
-    ).json()["id"]
+    )
+    try:
+      return r.json()["id"]
+    except Exception as e:
+      raise errors.OperationFailedError("Failed to send message", r.content) from e
   def buildReply(self, target, message):
     '''Convenience function for making a reply.
 
@@ -336,13 +359,13 @@ class Room:
       "Editing message {0} to: {1}"
         .format(target, newMessage)
     )
-    self.processTooFast(
+    r = self.processTooFast(
       lambda: self.session.post(
         "https://chat.stackexchange.com/messages/{}"
           .format(target),
         data = {
           "text": newMessage,
-          "fkey": self.fkey
+          "fkey": self._fkey
         },
         headers = {
           'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -351,6 +374,8 @@ class Room:
       ),
       handleTooFast
     )
+    if r.text != '"ok"':
+      raise errors.OperationFailedError("Failed to edit message", r.content)
   def delete(self, id, handleTooFast = True):
     '''Delete a message.
 
@@ -363,12 +388,12 @@ class Room:
       "Deleting message {}"
         .format(id)
     )
-    self.processTooFast(
+    r = self.processTooFast(
       lambda: self.session.post(
         "https://chat.stackexchange.com/messages/{}/delete"
           .format(id),
         data = {
-          "fkey": self.fkey
+          "fkey": self._fkey
         },
         headers = {
           'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -377,8 +402,10 @@ class Room:
       ),
       handleTooFast
     )
+    if r.text != '"ok"':
+      raise errors.OperationFailedError("Failed to delete message", r.content)
   def star(self, id, handleTooFast = True):
-    '''Star a message.
+    '''Toggle the starred status of a message.
 
         :param id: The message ID to star.
         :type id: int
@@ -389,12 +416,12 @@ class Room:
       "Starring {0}"
         .format(id)
     )
-    self.processTooFast(
+    r = self.processTooFast(
       lambda: self.session.post(
         "https://chat.stackexchange.com/messages/{}/star"
           .format(id),
         data = {
-          "fkey": self.fkey
+          "fkey": self._fkey
         },
         headers = {
           'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -403,6 +430,8 @@ class Room:
       ),
       handleTooFast
     )
+    if r.text != '"ok"':
+      raise errors.OperationFailedError("Failed to star message", r.content)
   def pin(self, id, handleTooFast = True):
     '''Pin a message.
 
@@ -415,12 +444,12 @@ class Room:
       "Pinning {0}"
         .format(id)
     )
-    self.processTooFast(
+    r = self.processTooFast(
       lambda: self.session.post(
         "https://chat.stackexchange.com/messages/{}/owner-star"
           .format(id),
         data = {
-          "fkey": self.fkey
+          "fkey": self._fkey
         },
         headers = {
           'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -429,6 +458,8 @@ class Room:
       ),
       handleTooFast
     )
+    if r.text != '"ok"':
+      raise errors.OperationFailedError("Failed to pin message", r.content)
   def unpin(self, id, handleTooFast = True):
     '''Unpin a message.
 
@@ -441,12 +472,12 @@ class Room:
       "Unpinning {0}"
         .format(id)
     )
-    self.processTooFast(
+    r = self.processTooFast(
       lambda: self.session.post(
         "https://chat.stackexchange.com/messages/{}/unowner-star"
           .format(id),
         data = {
-          "fkey": self.fkey
+          "fkey": self._fkey
         },
         headers = {
           'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -455,6 +486,8 @@ class Room:
       ),
       handleTooFast
     )
+    if r.text != '"ok"':
+      raise errors.OperationFailedError("Failed to unpin message", r.content)
   def clearStars(self, id, handleTooFast = True):
     '''Clear stars on a message.
 
@@ -467,12 +500,12 @@ class Room:
       "Clearing stars on {0}"
         .format(id)
     )
-    self.processTooFast(
+    r = self.processTooFast(
       lambda: self.session.post(
         "https://chat.stackexchange.com/messages/{}/unstar"
           .format(id),
         data = {
-          "fkey": self.fkey
+          "fkey": self._fkey
         },
         headers = {
           'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -481,6 +514,8 @@ class Room:
       ),
       handleTooFast
     )
+    if r.text != '"ok"':
+      raise errors.OperationFailedError("Failed to clear stars on message", r.content)
   def move(self, ids, target):
     '''Move a group of messages.
 
@@ -504,7 +539,7 @@ class Room:
       data = {
         "ids": ",".join([str(i) for i in ids]),
         "to": target,
-        "fkey": self.fkey
+        "fkey": self._fkey
       },
       headers = {
         'Referer': 'https://chat.stackexchange.com/rooms/{}'
@@ -526,6 +561,7 @@ class Bot:
     self.logger = logger if logger else logging.getLogger("Bot")
     
     self.session = requests.Session()
+    self.session.headers.update({"User-Agent": "Mozilla/5.0 (compatible; automated;) sechat/1.0.2 (unauthenticated; +http://pypi.org/project/sechat)"})
     self.fkey = None
     self.chatID = None
     self.userID = None
@@ -563,6 +599,7 @@ class Bot:
         else:
           l.debug("Success!")
           self.session.cookies.clear_expired_cookies()
+        f.close()
     if not "acct" in dict(self.session.cookies):
       self.logger.info("Logging in to " + host)
       self.logger.debug("Getting fkey...")
@@ -645,17 +682,20 @@ class Bot:
       ) from None
     self.logger.debug("Got chat fkey: " + self.fkey)
     self.logger.info("Logged in to chat successfully!")
+    self.session.headers.update({"User-Agent": "Mozilla/5.0 (compatible; automated;) sechat/1.0.2 (logged in as user %s; +http://pypi.org/project/sechat)" % self.userID})
 
-  def joinRoom(self, roomID):
+  def joinRoom(self, roomID, autoConnect = True):
     '''Join a room.
 
         :param roomID: The ID of the room to join.
-        :type roomID: str
+        :type roomID: int
 
         :return: The room instance.
         :rtype: sechat.Room
     '''
-    room = Room(self, roomID)
+    if roomID in self.rooms:
+      return self.rooms[roomID]
+    room = Room(self, roomID, autoConnect=autoConnect)
     self.rooms[roomID] = room
     return room
 
@@ -663,7 +703,7 @@ class Bot:
     '''Leave a room. The behavior of the corresponding Room instance afer this message is called is undefined.
 
         :param roomID: The room ID to leave.
-        :type roomID: str
+        :type roomID: int
         :param wait: If True, sechat guarantees that the room's thread will be stopped when the method returns.
         :type wait: bool
 
@@ -677,7 +717,7 @@ class Bot:
       self.rooms[roomID].halt(wait)
       self.rooms.pop(roomID)
     except KeyError:
-      raise ValueError("Not in room " + roomID) from None
+      raise ValueError("Not in room " + str(roomID)) from None
 
   def leaveAllRooms(self, wait = False):
     '''Leave all rooms. Works the same as leaveRoom but for all rooms.
@@ -711,4 +751,7 @@ class Bot:
       l.debug("Clearing cookies...")
       os.remove(self._COOKIEPATH)
       l.debug("Done!")
+    self.fkey = None
+    self.chatID = None
+    self.userID = None
     self.logger.info("Logged out successfully.")
