@@ -22,4 +22,7 @@ class TooFastError(RoomError):
   pass
 
 class NotAllowedError(RoomError):
-  '''Used when you try to do something you're not allowed to do'''
\ No newline at end of file
+  '''Used when you try to do something you're not allowed to do'''
+
+class OperationFailedError(RoomError):
+    '''Used when an operation fails (SEChat doesn't return "ok")'''
\ No newline at end of file
